@@ -12,7 +12,7 @@
  *
  * Also finds out IP of specified MAC
  *
- * $Id: arping.c,v 1.2 2002/11/03 19:58:21 marvin Exp $
+ * $Id: arping.c,v 1.6 2003/08/03 22:57:59 marvin Exp $
  */
 /*
  *  Copyright (C) 2000-2002 Thomas Habets <thomas@habets.pp.se>
@@ -31,7 +31,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
  */
-#include "config.h"
+//#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -66,7 +66,7 @@
 #define IP_ALEN 4
 #endif
 
-const float version = 2.0;
+const float version = 2.01;
 
 static libnet_t *libnet;
 
@@ -80,6 +80,7 @@ static int verbose = 0;
 static int finddup = 0;
 static unsigned int numsent = 0;
 static unsigned int numrecvd = 0;
+static int addr_must_be_same = 0;
 // RAWRAW is RAW|RRAW
 static enum { NORMAL,QUIET,RAW,RRAW,RAWRAW } display = NORMAL;
 static char *target = "huh? bug in arping?";
@@ -90,6 +91,75 @@ static char dstmac[ETH_ALEN];
 
 volatile int time_to_die = 0;
 
+
+const char *arping_lookupdev_default(u_int32_t srcip, u_int32_t dstip,
+				     char *ebuf)
+{
+	static const char *ifname;
+	if (!(ifname = pcap_lookupdev(ebuf))) {
+		return 0;
+	}
+	return ifname;
+}
+
+#if defined(FINDIF) && defined(linux)
+const char *arping_lookupdev(u_int32_t srcip, u_int32_t dstip, char *ebuf)
+{
+	FILE *f;
+	static const char buf[1024];
+	char buf1[1024];
+	char buf2[1024];
+	char *p,*p2;
+	int n;
+
+	libnet_host_lookup_r(dstip,0,buf2);
+	libnet_host_lookup_r(srcip,0,buf1);
+
+	/*
+	 * Construct and run command
+	 */
+	snprintf(buf, 1023, "/sbin/ip route get %s from %s 2>&1",
+		 buf2,buf1);
+	DEBUG(printf("%s\n",buf));
+	if (!(f = popen(buf, "r"))) {
+		goto failed;
+	}
+	if (0>(n = fread(buf, 1, sizeof(buf)-1, f))) {
+		pclose(f);
+		goto failed;
+	}
+	buf[n] = 0;
+	if (-1 == pclose(f)) {
+		perror("arping: pclose()");
+		goto failed;
+	}
+
+	/*
+	 * Parse out device
+	 */
+	p = strstr(buf, "dev ");
+	if (!p) {
+		goto failed;
+	}
+
+	p+=4;
+
+	p2 = strchr(p, ' ');
+	if (!p2) {
+		goto failed;
+	}
+	*p2 = 0;
+	return p;
+ failed:
+	return arping_lookupdev_default(srcip,dstip,ebuf);
+}
+#else
+const char *arping_lookupdev(u_int32_t srcip, u_int32_t dstip, char *ebuf)
+{
+	return arping_lookupdev_default(srcip,dstip,ebuf);
+}
+#endif
+
 /*
  *
  */
@@ -103,9 +173,9 @@ static void sigint(int i)
  */
 static void usage(int ret)
 {
-	printf("ARPing %1.2f BETA, by Thomas Habets <thomas@habets.pp.se>\n",
+	printf("ARPing %1.2f, by Thomas Habets <thomas@habets.pp.se>\n",
 	       version);
-	printf("usage: arping [ -qvrRd0bp ] [ -w <us> ] [ -S <host/ip> ] "
+	printf("usage: arping [ -0aAbdFpqrRv ] [ -w <us> ] [ -S <host/ip> ] "
 	       "[ -T <host/ip ]\n"
 	       "              [ -s <MAC> ] [ -t <MAC> ] [ -c <count> ] "
 	       "[ -i <interface> ]\n"
@@ -113,6 +183,38 @@ static void usage(int ret)
 	exit(ret);
 }
 
+/*
+ *
+ */
+static int is_mac_addr(const char *p)
+{
+	unsigned int n[6];
+	if(6==sscanf(p, "%2x%x.%2x%x.%2x%x",
+		     &n[0],&n[1],&n[2],&n[3],&n[4],&n[5])){
+		return 1;
+	}
+	return strchr(p, ':') ? 1 : 0;
+}
+
+/*
+ * lots of parms since C arrays suck
+ */
+static int get_mac_addr(const char *in,
+			unsigned int *n0,
+			unsigned int *n1,
+			unsigned int *n2,
+			unsigned int *n3,
+			unsigned int *n4,
+			unsigned int *n5)
+{
+	if (6 == sscanf(in, "%x:%x:%x:%x:%x:%x",n0,n1,n2,n3,n4,n5)) {
+		return 1;
+	} else if(6 == sscanf(in, "%2x%x.%2x%x.%2x%x",n0,n1,n2,n3,n4,n5)) {
+		return 1;
+	}
+	return 0;
+}
+
 /*
  * as always, the answer is 42
  *
@@ -215,7 +317,7 @@ static void pingmac_send(u_int8_t *srcmac, u_int8_t *dstmac,
 		printf("arping: sending packet\n");
 	}
 	if (-1 == (c = libnet_write(libnet))) {
-		fprintf(stderr, "libnet_write(): %s\n",
+		fprintf(stderr, "arping: libnet_write(): %s\n",
 			libnet_geterror(libnet));
 		sigint(0);
 	}
@@ -247,7 +349,7 @@ static void pingip_send(u_int8_t *srcmac, u_int8_t *dstmac,
 					  0,
 					  libnet,
 					  arp))) {
-		fprintf(stderr, "libnet_build_arp(): %s\n",
+		fprintf(stderr, "arping: libnet_build_arp(): %s\n",
 			libnet_geterror(libnet));
 		sigint(0);
 	}
@@ -258,7 +360,7 @@ static void pingip_send(u_int8_t *srcmac, u_int8_t *dstmac,
 					       0,
 					       libnet,
 					       eth))) {
-		fprintf(stderr, "libnet_build_ethernet(): %s\n",
+		fprintf(stderr, "arping: libnet_build_ethernet(): %s\n",
 			libnet_geterror(libnet));
 		sigint(0);
 	}
@@ -307,6 +409,11 @@ static void pingip_recv(const char *unused, struct pcap_pkthdr *h,
 		u_int32_t ip;
 		memcpy(&ip, (char*)harp + harp->ar_hln
 		       + LIBNET_ARP_H,4);
+		if (addr_must_be_same
+		    && (memcmp((u_char*)harp+sizeof(struct libnet_arp_hdr),
+			       dstmac, ETH_ALEN))) {
+			return;
+		}
 		if (dstip == ip) {
 			switch(display) {
 			case NORMAL: {
@@ -317,7 +424,7 @@ static void pingip_recv(const char *unused, struct pcap_pkthdr *h,
 					       (c<5)?':':' ');
 				}
 				
-				printf("(%s): index=%d time=%s\n",
+				printf("(%s): index=%d time=%s",
 				       libnet_addr2name4(ip,0),
 				       numrecvd,
 				       tv2str(&lastpacketsent,
@@ -330,20 +437,21 @@ static void pingip_recv(const char *unused, struct pcap_pkthdr *h,
 					printf("%.2x%c", heth->_802_3_shost[c],
 					       (c<5)?':':' ');
 				}
-				printf("%s\n", libnet_addr2name4(ip,0));
-				break;
-			case RAW:
-				printf("%s\n", libnet_addr2name4(ip,0));
+				printf("%s", libnet_addr2name4(ip,0));
 				break;
 			case RRAW:
+				printf("%s", libnet_addr2name4(ip,0));
+				break;
+			case RAW:
 				for (c = 0; c < 6; c++) {
-					printf("%.2x%c", heth->_802_3_shost[c],
-					       (c<5)?':':'\n');
+					printf("%.2x%s", heth->_802_3_shost[c],
+					       (c<5)?":":"");
 				}
 				break;
 			default:
 				fprintf(stderr, "arping: can't happen!\n");
 			}
+			printf(beep?"\a\n":"\n");
 			numrecvd++;
 		}
 	}
@@ -380,6 +488,13 @@ static void pingmac_recv(const char *unused, struct pcap_pkthdr *h,
 		 || !memcmp(dstmac, ethxmas, ETH_ALEN)))
 	    && !memcmp(heth->_802_3_dhost, srcmac, ETH_ALEN)) {
 /*		u_int8_t *cp = heth->_802_3_shost; */
+		if (addr_must_be_same) {
+			u_int32_t tmp;
+			memcpy(&tmp, &hip->ip_src, 4);
+			if (dstip != tmp) {
+				return;
+			}
+		}
 		switch(display) {
 		case QUIET:
 			break;
@@ -391,18 +506,18 @@ static void pingmac_recv(const char *unused, struct pcap_pkthdr *h,
 				printf("%.2x%c", heth->_802_3_shost[c],
 				       (c<5)?':':')');
 			}
-			printf(": icmp_seq=%d time=%s\n",
+			printf(": icmp_seq=%d time=%s",
 			       hicmp->icmp_seq,tv2str(&lastpacketsent,
 						      &arrival,buf));
 			break; }
 		case RAW:
-			printf("%s\n",
+			printf("%s",
 			       libnet_addr2name4(hip->ip_src.s_addr, 0));
 			break;
 		case RRAW:
 			for (c = 0; c < 6; c++) {
-				printf("%.2x%c", heth->_802_3_shost[c],
-				       (c<5)?':':'\n');
+				printf("%.2x%s", heth->_802_3_shost[c],
+				       (c<5)?":":"");
 			}
 			break;
 		case RAWRAW:
@@ -410,13 +525,14 @@ static void pingmac_recv(const char *unused, struct pcap_pkthdr *h,
 				printf("%.2x%c", heth->_802_3_shost[c],
 				       (c<5)?':':' ');
 			}
-			printf("%s\n",
+			printf("%s",
 			       libnet_addr2name4(hip->ip_src.s_addr, 0));
 			break;
 		default:
 			fprintf(stderr, "arping: can't-happen-bug\n");
 			sigint(0);
 		}
+		printf(beep?"\a\n":"\n");
 		numrecvd++;
 	}
 }
@@ -531,6 +647,7 @@ int main(int argc, char **argv)
 	char *parm;
 	int c;
 	unsigned int maxcount = -1;
+	int dont_use_arping_lookupdev=0;
 	struct bpf_program bp;
 	pcap_t *pcap;
 	static enum { NONE, PINGMAC, PINGIP } mode = NONE;
@@ -543,13 +660,34 @@ int main(int argc, char **argv)
 	memset(dstmac, 0xff, ETH_ALEN);
 	memset(ethxmas, 0xff, ETH_ALEN);
 
-	while ((c = getopt(argc, argv, "0bdS:T:Bvhi:rRc:qs:t:paw:")) != EOF) {
+	while (EOF!=(c = getopt(argc, argv, "0aAbBc:dFhi:pqrRs:S:t:T:vw:"))) {
 		switch(c) {
+		case '0':
+			srcip = 0;
+			srcip_given = 1;
+			break;
 		case 'a':
 			beep = 1;
 			break;
-		case 'v':
-			verbose++;
+		case 'A':
+			addr_must_be_same = 1;
+			break;
+		case 'b':
+			srcip = 0xffffffff;
+			srcip_given = 1;
+			break;
+		case 'B':
+			dstip = 0xffffffff;
+			dstip_given = 1;
+			break;
+		case 'c':
+			maxcount = atoi(optarg);
+			break;
+		case 'd':
+			finddup = 1;
+			break;
+		case 'F':
+			dont_use_arping_lookupdev=1;
 			break;
 		case 'h':
 			usage(0);
@@ -563,21 +701,32 @@ int main(int argc, char **argv)
 			}
 			ifname = optarg;
 			break;
+		case 'p':
+			promisc = 1;
+			break;
+		case 'q':
+			display = QUIET;
+			break;
 		case 'r':
 			display = (display==RRAW)?RAWRAW:RAW;
 			break;
 		case 'R':
 			display = (display==RAW)?RAWRAW:RRAW;
 			break;
-		case 'q':
-			display = QUIET;
-			break;
-		case 'c':
-			maxcount = atoi(optarg);
-			break;
-		case 'd':
-			finddup = 1;
+		case 's': {// spoof source MAC
+			unsigned int n[6];
+			if (!get_mac_addr(optarg,
+					  &n[0],&n[1],&n[2],
+					  &n[3],&n[4],&n[5])){
+				fprintf(stderr, "arping: Weird MAC addr %s\n",
+					optarg);
+				exit(1);
+			}
+			for (c = 0; c < 6; c++) {
+				srcmac[c] = n[c] & 0xff;
+			}
 			break;
+		}
 		case 'S': // set source IP, may be null for don't-know
 			if (-1 == (srcip = libnet_name2addr4(libnet,
 							     optarg,
@@ -589,6 +738,26 @@ int main(int argc, char **argv)
 			}
 			srcip_given = 1;
 			break;
+		case 't': { // set taget mac
+			unsigned int n[6];
+			if (mode == PINGMAC) {
+				fprintf(stderr, "arping: -t can only be used "
+					"in IP ping mode\n");
+				exit(1);
+			}
+			if (!get_mac_addr(optarg,
+					  &n[0],&n[1],&n[2],
+					  &n[3],&n[4],&n[5])){
+				fprintf(stderr, "Illegal MAC addr %s\n",
+					optarg);
+				exit(1);
+			}
+			for (c = 0; c < 6; c++) {
+				dstmac[c] = n[c] & 0xff;
+			}
+			break;
+			mode = PINGIP;
+		}
 		case 'T': // set destination IP
 			if (mode == PINGIP) {
 				fprintf(stderr, "arping: -T can only be used "
@@ -605,66 +774,8 @@ int main(int argc, char **argv)
 			}
 			mode = PINGMAC;
 			break;
-		case 'b':
-			srcip = 0xffffffff;
-			srcip_given = 1;
-			break;
-		case 'B':
-			dstip = 0xffffffff;
-			dstip_given = 1;
-			break;
-		case '0':
-			srcip = 0;
-			srcip_given = 1;
-			break;
-		case 's': {// spoof source MAC
-			unsigned int n[6];
-			   
-			if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
-				   &n[0],
-				   &n[1],
-				   &n[2],
-				   &n[3],
-				   &n[4],
-				   &n[5]
-				    ) != 6) {
-				fprintf(stderr, "arping: Weird MAC addr %s\n",
-					optarg);
-				exit(1);
-			}
-			for (c = 0; c < 6; c++) {
-				srcmac[c] = n[c] & 0xff;
-			}
-			break;
-		}
-		case 't': { // set taget mac
-			unsigned int n[6];
-			if (mode == PINGMAC) {
-				fprintf(stderr, "arping: -t can only be used "
-					"in IP ping mode\n");
-				exit(1);
-			}
-			
-			if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
-				   &n[0],
-				   &n[1],
-				   &n[2],
-				   &n[3],
-				   &n[4],
-				   &n[5]
-				    ) != 6) {
-				fprintf(stderr, "Illegal MAC addr %s\n",
-					optarg);
-				exit(1);
-			}
-			for (c = 0; c < 6; c++) {
-				dstmac[c] = n[c] & 0xff;
-			}
-			break;
-			mode = PINGIP;
-		}
-		case 'p':
-			promisc = 1;
+		case 'v':
+			verbose++;
 			break;
 		case 'w':
 			packetwait = (unsigned)atoi(optarg);
@@ -697,8 +808,9 @@ int main(int argc, char **argv)
 	 * Handle dstip_given instead of ip address after parms (-B really)
 	 */
 	if (mode == NONE) {
+		unsigned char n[6];
 		if (optind + 1 == argc) {
-			mode = strchr(parm, ':')?PINGMAC:PINGIP;
+			mode = is_mac_addr(parm)?PINGMAC:PINGIP;
 		} else if (dstip_given) {
 			mode = PINGIP;
 			parm = strdup(libnet_addr2name4(dstip,
@@ -714,7 +826,7 @@ int main(int argc, char **argv)
 	 * Make sure dstip and parm like eachother
 	 */
 	if (mode == PINGIP && (!dstip_given)) {
-		if (strchr(parm, ':')) {
+		if (is_mac_addr(parm)) {
 			fprintf(stderr, "arping: Options given only apply to "
 				"IP ping, but MAC address given as argument"
 				"\n");
@@ -733,25 +845,21 @@ int main(int argc, char **argv)
 	 * parse parm into dstmac
 	 */
 	if (mode == PINGMAC) {
-		unsigned char n[6];
+		unsigned int n[6];
 		if (optind + 1 != argc) {
 			usage(1);
 		}
-		if (!strchr(parm, ':')) {
+		if (!is_mac_addr(parm)) {
 			fprintf(stderr, "arping: Options given only apply to "
 				"MAC ping, but no MAC address given as "
 				"argument\n");
 			exit(1);
 		}
-		if (sscanf(argv[optind], "%x:%x:%x:%x:%x:%x", 
-			   &n[0],
-			   &n[1],
-			   &n[2],
-			   &n[3],
-			   &n[4],
-			   &n[5]
-			   ) != 6) {
-			fprintf(stderr, "Illegal mac addr %s\n", argv[optind]);
+		if (!get_mac_addr(argv[optind],
+				  &n[0],&n[1],&n[2],
+				  &n[3],&n[4],&n[5])) {
+			fprintf(stderr, "arping: Illegal mac addr %s\n",
+				argv[optind]);
 			return 1;
 		}
 		for (c = 0; c < 6; c++) {
@@ -769,12 +877,19 @@ int main(int argc, char **argv)
 	 * right one.
 	 */
 	if (!ifname) {
-		if (!(ifname = pcap_lookupdev(ebuf))) {
-			fprintf(stderr, "arping: pcap_lookupdev(): %s\n",ebuf);
+		if (dont_use_arping_lookupdev) {
+			ifname = arping_lookupdev_default(srcip,dstip,ebuf);
+		} else {
+			ifname = arping_lookupdev(srcip,dstip,ebuf);
+		}
+		if (!ifname) {
+			fprintf(stderr, "arping: arping_lookupdev(): %s\n",
+				ebuf);
 			exit(1);
 		}
 		// FIXME: check for other probably-not interfaces
-		if (!strncmp(ifname, "ipsec", 5)) {
+		if (!strcmp(ifname, "ipsec")
+		    || !strcmp(ifname,"lo")) {
 			fprintf(stderr, "arping: Um.. %s looks like the wrong "
 				"interface to use. Is it? "
 				"(-i switch)\n", ifname);
@@ -824,7 +939,7 @@ int main(int argc, char **argv)
 	 */
 	if (!srcmac_given) {
 		if (!(cp = (char*)libnet_get_hwaddr(libnet))) {
-			fprintf(stderr, "libnet_get_hwaddr(): %s\n",
+			fprintf(stderr, "arping: libnet_get_hwaddr(): %s\n",
 				libnet_geterror(libnet));
 			exit(1);
 		}
