@@ -1,3 +1,9 @@
+/*
+ * NOTE: This is arping 1.x, arping 2.x is in arping-2/
+ */
+
+
+
 /*
  * arping
  *
@@ -12,50 +18,56 @@
  *
  * Also finds out IP of specified MAC
  *
- * $Id: arping.c,v 1.60 2003/04/07 18:05:59 marvin Exp $
+ * $Id: arping.c,v 1.64 2003/08/03 23:06:58 marvin Exp $
  */
 /*
- *  Copyright (C) 2000-2002 Thomas Habets <thomas@habets.pp.se>
+ *  Copyright (C) 2000-2003 Thomas Habets <thomas@habets.pp.se>
  *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU General Public
- *  License as published by the Free Software Foundation; either
- *  version 2 of the License, or (at your option) any later version.
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation; either version 2 of the License, or
+ *   (at your option) any later version.
  *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  General Public License for more details.
+ *   This program is distributed in the hope that it will be useful,
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *   GNU General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ *   You should have received a copy of the GNU General Public License
+ *   along with this program; if not, write to the Free Software
+ *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 /*
  * Note to self:
- *  Test checklist:
+ *  Test checklist:  (cmac = mac in cisco format (0000.0000.0000)
  *    command                   expected response
  *    arping host               pongs
  *    arping -a host            audiable pongs
  *    arping mac                pongs
+ *    arping cmac               pongs
  *    arping -a mac             audiable pongs
  *    arping -A host            nothing
  *    arping -A mac             nothing
- *    arping -t mac -A host     nothing
- *    arping -T ip -A mac       nothing
+ *    arping -t cmac -A host    pongs
+ *    arping -t mac -A host     pongs
+ *    arping -T ip -A mac       pongs
+ *    arping -T ip -A cmac      pongs
+ *    arping -T bcast mac       pongs    # bcast of current net
  *    arping -r host            mac
  *    arping -R host            ip
  *    arping -r mac             ip
  *    arping -R mac             mac
  *    arping -rR mac            mac ip
+ *    arping -rR ip             mac ip
  *    ./arping-scan-net.sh mac  ip
  * 
  * Arch checklist:
- *   Solaris/sparc
- *   NetBSD/alpha      (is libnet or I unaligned? -- libnet I think)
- *   OpenBSD/sparc64   (libnet a bit buggy)
- *   Linux/x86
+ *   Linux/x86         test with debian package libnet0-dev and libnet1-dev
  *   Linux/sparc
+ *   Linux/hppa
+ *   Solaris/sparc
+ *   NetBSD/alpha      (is libnet or arping unaligned? -- libnet I think)
+ *   OpenBSD/sparc64   (libnet a bit buggy here)
  *
  */
 #include <stdio.h>
@@ -105,7 +117,7 @@
 #define DEBUG(a)
 #endif
 
-const float version = 1.07;
+const float version = 1.08;
 
 struct ether_addr *mymac;
 static u_char eth_xmas[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
@@ -142,6 +154,75 @@ static unsigned int nullip = 0;
 static unsigned int is_promisc = 0;
 static unsigned int addr_must_be_same = 0;
 
+
+const char *arping_lookupdev_default(u_int32_t srcip, u_int32_t dstip,
+				     char *ebuf)
+{
+	static const char *ifname;
+	if (!(ifname = pcap_lookupdev(ebuf))) {
+		return 0;
+	}
+	return ifname;
+}
+
+#if defined(FINDIF) && defined(linux)
+const char *arping_lookupdev(u_int32_t srcip, u_int32_t dstip, char *ebuf)
+{
+	FILE *f;
+	static char buf[1024];
+	char buf1[1024];
+	char buf2[1024];
+	char *p,*p2;
+	int n;
+
+	libnet_host_lookup_r(dstip,0,buf2);
+	libnet_host_lookup_r(srcip,0,buf1);
+
+	/*
+	 * Construct and run command
+	 */
+	snprintf(buf, 1023, "/sbin/ip route get %s from %s 2>&1",
+		 buf2,buf1);
+	DEBUG(printf("%s\n",buf));
+	if (!(f = popen(buf, "r"))) {
+		goto failed;
+	}
+	if (0 > (n = fread(buf, 1, sizeof(buf)-1, f))) {
+		pclose(f);
+		goto failed;
+	}
+	buf[n] = 0;
+	if (-1 == pclose(f)) {
+		perror("arping: pclose()");
+		goto failed;
+	}
+
+	/*
+	 * Parse out device
+	 */
+	p = strstr(buf, "dev ");
+	if (!p) {
+		goto failed;
+	}
+
+	p+=4;
+
+	p2 = strchr(p, ' ');
+	if (!p2) {
+		goto failed;
+	}
+	*p2 = 0;
+	return p;
+ failed:
+	return arping_lookupdev_default(srcip,dstip,ebuf);
+}
+#else
+const char *arping_lookupdev(u_int32_t srcip, u_int32_t dstip, char *ebuf)
+{
+	return arping_lookupdev_default(srcip,dstip,ebuf);
+}
+#endif
+
 static void sigint(int i)
 {
 	DEBUG(printf("sigint()\n"));
@@ -170,8 +251,8 @@ static void usage(int ret)
 {
 	printf("ARPing %1.2f, by Thomas Habets <thomas@habets.pp.se>\n",
 	       version);
-	printf("usage: arping [ -qvrRd0bpAa ] [ -S <host/ip> ] [ -T <host/ip ]"
-	       " [ -s <MAC> ]\n"
+	printf("usage: arping [ -0aAbdFpqrRv ] [ -S <host/ip> ] "
+	       "[ -T <host/ip ] [ -s <MAC> ]\n"
 	       "              [ -t <MAC> ] [ -c <count> ] [ -i <interface> ] "
 	       "<host/ip/MAC | -B>\n");
 	exit(ret);
@@ -187,7 +268,7 @@ static void alasend(int i)
 	if (searchmac) {
 		libnet_build_icmp_echo(ICMP_ECHO,      /* type */ 
 				       0,              /* code */ 
-				       (short)random(),           /* id */ 
+				       (short)rand(),           /* id */ 
 				       htons(numsent-1), /* seq */ 
 				       NULL,           /* pointer to payload */
 				       0,              /* payload length */ 
@@ -375,8 +456,10 @@ static void handlepacket(const char *unused, struct pcap_pkthdr *h,
 					}
 					if (!rawoutput) {
 						printf(" (%s): index=%d time=%s",
-						       libnet_host_lookup(ip, 0),
-						       numrecvd, tvtoda(&lastpacketsent, &recvtime));
+						       libnet_host_lookup(ip,0),
+						       numrecvd,
+						       tvtoda(&lastpacketsent,
+							      &recvtime));
 					}
 					if (beep) {
 						printf("\a");
@@ -407,6 +490,8 @@ int main(int argc, char **argv)
 	struct bpf_program bp;
 	char must_be_pingip = 0;
 	char have_eth_source = 0;
+	unsigned int n[6];
+	int dont_use_arping_lookupdev=0;
 	
 	DEBUG(printf("main()\n"));
 
@@ -415,7 +500,7 @@ int main(int argc, char **argv)
 
 	memcpy(eth_target, eth_xmas, ETH_ALEN);
 
-	while ((c = getopt(argc, argv, "0bdS:T:Bvhi:rRc:qs:t:paA")) != EOF) {
+	while ((c = getopt(argc, argv, "aAbBc:dF0S:T:hi:rRqs:t:pv")) != EOF) {
 		switch (c) {
 		case 'A':
 			addr_must_be_same = 1;
@@ -423,6 +508,19 @@ int main(int argc, char **argv)
 		case 'a':
 			beep = 1;
 			break;
+		case 'c':
+			maxcount = atoi(optarg);
+			break;
+		case 'd':
+			finddup = 1;
+			break;	
+		case 'F':
+#if !defined(FINDIF)
+			fprintf(stderr, "arping: find-interface support not "
+				"compiled in\n");
+#endif
+			dont_use_arping_lookupdev=1;
+			break;
 		case 'v':
 			verbose++;
 			break;
@@ -447,16 +545,9 @@ int main(int argc, char **argv)
 		case 'q':
 			quiet = rawoutput = 1;
 			break;
-		case 'c':
-			maxcount = atoi(optarg);
-			break;
-		case 'd':
-			finddup = 1;
-			break;	
 		case 'S':
-			if ((myip = libnet_name_resolve(optarg,
-							LIBNET_RESOLVE))
-			    == -1) {
+			if (-1==(myip = libnet_name_resolve(optarg,
+							    LIBNET_RESOLVE))){
 				fprintf(stderr,"arping: Can't to resolve %s\n",
 					optarg);
 				exit(1);
@@ -466,8 +557,12 @@ int main(int argc, char **argv)
 			}
 			break;
 		case 'T': // destination IP in mac ping (default: 0xffffffff)
-			dip = libnet_name_resolve(optarg,
-						  LIBNET_RESOLVE);
+			if (-1 == (dip = libnet_name_resolve(optarg,
+							     LIBNET_RESOLVE))){
+				fprintf(stderr, "arping: can't resolve: %s\n",
+					optarg);
+				exit(1);
+			}
 			searchmac = 1;
 			break;
 		case 'b':
@@ -480,47 +575,62 @@ int main(int argc, char **argv)
 			nullip = 1;
 			break;
 		case 's': // spoofed source MAC
-			{
-                           unsigned int n[6];
-			   
-			   if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
-				      &n[0],
-				      &n[1],
-				      &n[2],
-				      &n[3],
-				      &n[4],
-				      &n[5]
-				   ) != 6) {
-				   fprintf(stderr, "Illegal MAC addr %s\n",
-					   optarg);
-				   exit(1);
-			   }
-			   for (c = 0; c < 6; c++) {
-				   eth_source[c] = n[c] & 0xff;
-			   }
-			   have_eth_source = 1;
+			if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
+				   &n[0],
+				   &n[1],
+				   &n[2],
+				   &n[3],
+				   &n[4],
+				   &n[5]
+				    ) == 6) {
+				;
+			} else if(sscanf(optarg, "%2x%x.%2x%x.%2x%x",
+					  &n[0],
+					  &n[1],
+					  &n[2],
+					  &n[3],
+					  &n[4],
+					  &n[5]
+					   ) == 6) {
+				   ;
+			} else {
+				fprintf(stderr, "arping: Illegal MAC addr "
+					"%s\n", optarg);
+				exit(1);
 			}
+			for (c = 0; c < 6; c++) {
+				eth_source[c] = n[c] & 0xff;
+			}
+			have_eth_source = 1;
 			break;
 		case 't':
 			must_be_pingip = 1;
-			{
-                           unsigned int n[6];
-			   
-			   if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
-				      &n[0],
-				      &n[1],
-				      &n[2],
-				      &n[3],
-				      &n[4],
-				      &n[5]
-				   ) != 6) {
-				   fprintf(stderr, "Illegal MAC addr %s\n",
-					   optarg);
-				   exit(1);
-			   }
-			   for (c = 0; c < 6; c++) {
-				   eth_target[c] = n[c] & 0xff;
-			   }
+			if (sscanf(optarg, "%x:%x:%x:%x:%x:%x",
+				   &n[0],
+				   &n[1],
+				   &n[2],
+				   &n[3],
+				   &n[4],
+				   &n[5]
+				    ) == 6) {
+				;
+			} else if(sscanf(optarg, "%2x%x.%2x%x.%2x%x",
+					 &n[0],
+					 &n[1],
+					 &n[2],
+					 &n[3],
+					 &n[4],
+					 &n[5]
+					  ) == 6) {
+				;    
+			} else {
+				fprintf(stderr,"arping: Illegal MAC addr %s\n",
+					optarg);
+				exit(1);
+			}
+			
+			for (c = 0; c < 6; c++) {
+				eth_target[c] = n[c] & 0xff;
 			}
 			break;
 		case 'p':
@@ -545,41 +655,52 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	if (searchmac || (!dip && strchr(argv[optind], ':'))) {
-		unsigned int n[6];
-		if (!dip) {
-			dip = ip_xmas;
-		}
-
-		if (must_be_pingip) {
-			fprintf(stderr, "Specified switch can't be used in "
-				"MAC-ping mode\n");
-			exit(1);
-		}
-
-		if (sscanf(argv[optind], "%x:%x:%x:%x:%x:%x", 
-			   &n[0],
-			   &n[1],
-			   &n[2],
-			   &n[3],
-			   &n[4],
-			   &n[5]
-			   ) != 6) {
-			fprintf(stderr, "Illegal mac addr %s\n", argv[optind]);
-			return 1;
-		}
-		for (c = 0; c < 6; c++) {
-			eth_target[c] = n[c] & 0xff;
-		}
+	if (sscanf(argv[optind], "%x:%x:%x:%x:%x:%x", 
+		   &n[0],
+		   &n[1],
+		   &n[2],
+		   &n[3],
+		   &n[4],
+		   &n[5]
+		   ) == 6) {
+		searchmac = 1;
+	} else if(sscanf(argv[optind], "%2x%x.%2x%x.%2x%x",
+		   &n[0],
+		   &n[1],
+		   &n[2],
+		   &n[3],
+		   &n[4],
+		   &n[5]
+		   ) == 6) {
 		searchmac = 1;
 	} else if (!dip) {
-		
-		if (-1 == (dip = libnet_name_resolve((u_char*)argv[optind],
-						     LIBNET_RESOLVE))) {
+		if (-1 == (dip=libnet_name_resolve((u_char*)argv[optind],
+						   LIBNET_RESOLVE))) {
 			fprintf(stderr, "arping: Can't resolve %s\n",
 				argv[optind]);
 			exit(1);
 		}
+	} else if (must_be_pingip) {
+		fprintf(stderr, "arping: Illegal IP %s\n", argv[optind]);
+		exit(1);
+	} else {
+		fprintf(stderr, "arping: Illegal mac %s\n", argv[optind]);
+		exit(1);
+	}
+	if (searchmac && !dip) {
+		dip = ip_xmas;
+	}
+
+	if (searchmac && must_be_pingip) {
+		fprintf(stderr, "arping: Specified switch can't be used in "
+			"MAC-ping mode\n");
+		exit(1);
+	}
+
+	if (searchmac) {
+		for (c = 0; c < 6; c++) {
+			eth_target[c] = n[c] & 0xff;
+		}
 	}
 	if (finddup && maxcount == -1) {
 		maxcount = 3;
@@ -588,12 +709,18 @@ int main(int argc, char **argv)
 	 * libnet init
 	 */
 	if (!ifname) {
-		if (!(ifname = pcap_lookupdev(ebuf))) {
-			fprintf(stderr, "pcap_lookupdev(): %s\n", ebuf);
+		if (dont_use_arping_lookupdev) {
+			ifname = arping_lookupdev_default(myip,dip,ebuf);
+		} else {
+			ifname = arping_lookupdev(myip,dip,ebuf);
+		}
+		if (!ifname) {
+			fprintf(stderr,"arping_lookupdev(): %s\n", ebuf);
 			exit(1);
 		}
 		// FIXME: check for other probably-not interfaces
-		if (!strncmp(ifname, "ipsec", 5)) {
+		if (!strcmp(ifname, "ipsec")
+		    || !strcmp(ifname, "lo")) {
 			fprintf(stderr, "arping: Um.. %s looks like the wrong "
 				"interface to use. Is it? "
 				"(-i switch)\n", ifname);
